@@ -7,7 +7,7 @@ int lunzip(unsigned char *inbuf, long in_len,
 	   long (*flush)(void*, unsigned long),
 	   unsigned char *outbuf,
 	   long *in_posp,
-	   void (*error)(char *x));
+	   void (*error)(const char *x));
 
 /* This internal function is required because the decompress_fn above
  * (see include/linux/decompress/generic.h) should have an out_size
@@ -19,5 +19,5 @@ int __lunzip(unsigned char *inbuf, long in_len,
 	     long (*flush)(void*, unsigned long),
 	     unsigned char *outbuf, long out_size,
 	     long *in_posp, long *out_posp,
-	     void (*error)(char *x));
+	     void (*error)(const char *x));
 #endif
