@@ -1,7 +1,7 @@
 /*
  * LZIP decompressor
  *
- * Copyright (C) 2016-2025 Antonio Diaz Diaz.
+ * Copyright (C) 2016-2026 Antonio Diaz Diaz.
  *
  * Licensed under GPLv2 or later, see file LICENSE in this source tree.
  */
@@ -317,7 +317,8 @@ static inline bool Rd_init(struct Range_decoder * const rdec,
 	rdec->partial_member_pos = 0;
 	rdec->buffer_given = (inbuf && in_len > 0);
 	rdec->buffer_size = rdec->buffer_given ? in_len : 16384;
-	rdec->buffer = rdec->buffer_given ? inbuf : malloc(rdec->buffer_size);
+	rdec->buffer = rdec->buffer_given ?
+	               inbuf : (uint8_t *)malloc(rdec->buffer_size);
 	if (!rdec->buffer)
 		return false;
 	rdec->pos = 0;
@@ -639,7 +640,8 @@ static inline bool LZd_init(struct LZ_decoder * const d,
 	d->buffer_given = (outbuf && out_size > 0);
 	d->buffer_size = d->buffer_given ? (unsigned long)out_size : dict_size;
 	d->dictionary_size = min_t(unsigned long, d->buffer_size, dict_size);
-	d->buffer = d->buffer_given ? outbuf : large_malloc(d->buffer_size);
+	d->buffer = d->buffer_given ?
+	            outbuf : (uint8_t *)large_malloc(d->buffer_size);
 	if (!d->buffer)
 		return false;
 	d->pos = 0;
@@ -762,22 +764,18 @@ static int LZd_decode_member(struct LZ_decoder * const d)
 						LZd_flush_data(d);
 						if (d->write_error)
 							return LZIP_WRITE_ERROR;
-						if (len == min_match_len) {	/* End Of Stream marker */
-							if (LZd_check_trailer(d))
-								return 0;
-							else
-								return LZIP_BAD_CRC;
-						}
+						if (len == min_match_len)	/* End Of Stream marker */
+							return LZd_check_trailer(d) ? 0 : LZIP_BAD_CRC;
 						return LZIP_BAD_DATA;	/* unknown marker */
 					}
 				}
 			}
-			state = St_set_match(state);
 			if (rep0 >= d->dictionary_size ||
 			   (rep0 >= d->pos && !d->pos_wrapped)) {
 				LZd_flush_data(d);
 				return LZIP_BAD_DATA;
 			}
+			state = St_set_match(state);
 		}
 		LZd_copy_block(d, rep0, len);
 	}
@@ -842,7 +840,7 @@ int lzip_decompress(unsigned char *inbuf, long in_len,
 		}
 
 		if (!decoder)
-			decoder = malloc(sizeof *decoder);
+			decoder = (struct LZ_decoder *)malloc(sizeof *decoder);
 		if (!decoder || !LZd_init(decoder, &rdec, dictionary_size,
 					outptr, out_size, flush)) {
 			retval = LZIP_OOM_OUTBUF;
