@@ -1,5 +1,5 @@
 /* Xlunzip - Test tool for the lzip_decompress linux module
-   Copyright (C) 2016-2025 Antonio Diaz Diaz.
+   Copyright (C) 2016-2026 Antonio Diaz Diaz.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -80,7 +80,7 @@ static void show_error( const char * const msg, const int errcode,
                         const bool help );
 
 static const char * const program_name = "xlunzip";
-static const char * const program_year = "2025";
+static const char * const program_year = "2026";
 static const char * invocation_name = "xlunzip";	/* default value */
 
 static const struct { const char * from; const char * to; } known_extensions[] = {
@@ -97,50 +97,52 @@ static bool delete_output_on_interrupt = false;
 
 static void show_help( void )
   {
-  printf( "Xlunzip is a test tool for the lzip decompression code of my lzip patch for\n"
-          "linux. Xlunzip is similar to lunzip, but it uses the lzip_decompress linux\n"
-          "module as a backend. Xlunzip tests the module for stream, buffer-to-buffer,\n"
-          "and mixed decompression modes, including in-place decompression (using the\n"
-          "same buffer for input and output). You can use xlunzip to check that the\n"
-          "module produces correct results when decompressing single member files,\n"
-          "multimember files, or the concatenation of two or more compressed files.\n"
-          "Xlunzip can be used with unzcrash to test the robustness of the module to\n"
-          "the decompression of corrupted data.\n"
-          "\nThe distributed index feature of the lzip format allows xlunzip to\n"
-          "decompress concatenated files in place. This can't be guaranteed to work\n"
-          "with formats like gzip or bzip2 because they can't detect whether a high\n"
-          "compression ratio in the first members of the multimember data is being\n"
-          "masked by a low compression ratio in the last members.\n"
-          "\nUsage: %s [options] [files]\n", invocation_name );
-  printf( "\nOptions:\n"
-          "  -h, --help                 display this help and exit\n"
-          "  -V, --version              output version information and exit\n"
-          "  -c, --stdout               write to standard output, keep input files\n"
-          "  -d, --decompress           decompress (this is the default)\n"
-          "  -f, --force                overwrite existing output files\n"
-          "  -I, --in-place             decompress or test using only one buffer\n"
-          "  -k, --keep                 keep (don't delete) input files\n"
-          "  -o, --output=<file>        write to <file>, keep input files\n"
-          "  -q, --quiet                suppress all messages\n"
-          "  -t, --test                 test compressed file integrity\n"
-          "  -v, --verbose              be verbose (a 2nd -v gives more)\n"
-          "These options are ignored when --in-place is in effect:\n"
-          "      --insize[=<size>]      pre-allocate and fill inbuf [default 16 KiB]\n"
-          "      --outsize[=<size>]     pre-allocate outbuf [default 512 MiB]\n"
-          "      --nofill               do not pass a fill function; requires --insize\n"
-          "      --noflush              do not pass a flush function; requires --outsize\n"
-          "\nIf no file names are given, or if a file is '-', xlunzip decompresses\n"
-          "from standard input to standard output.\n"
-          "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
-          "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
-          "\nTo extract all the files from archive 'foo.tar.lz', use the commands\n"
-          "'tar -xf foo.tar.lz' or 'xlunzip -cd foo.tar.lz | tar -xf -'.\n"
-          "\nExit status: 0 for a normal exit, 1 for environmental problems\n"
-          "(file not found, invalid command-line options, I/O errors, etc), 2 to\n"
-          "indicate a corrupt or invalid input file, 3 for an internal consistency\n"
-          "error (e.g., bug) which caused xlunzip to panic.\n"
-          "\nReport bugs to lzip-bug@nongnu.org\n"
-          "Xlunzip home page: http://www.nongnu.org/lzip/xlunzip.html\n" );
+  fputs(
+    "Xlunzip is a test tool for the lzip decompression code of my lzip patch for\n"
+    "linux. Xlunzip is similar to lunzip, but it uses the lzip_decompress linux\n"
+    "module as a backend. Xlunzip tests the module for stream, buffer-to-buffer,\n"
+    "and mixed decompression modes, including in-place decompression (using the\n"
+    "same buffer for input and output). You can use xlunzip to check that the\n"
+    "module produces correct results when decompressing single member files,\n"
+    "multimember files, or the concatenation of two or more compressed files.\n"
+    "Xlunzip can be used with unzcrash to test the robustness of the module to\n"
+    "the decompression of corrupted data.\n"
+    "\nThe distributed index feature of the lzip format allows xlunzip to\n"
+    "decompress concatenated files in place. This can't be guaranteed to work\n"
+    "with formats like gzip or bzip2 because they can't detect whether a high\n"
+    "compression ratio in the first members of the multimember data is being\n"
+    "masked by a low compression ratio in the last members.\n", stdout );
+  printf( "\nUsage: %s [options] [files]\n", invocation_name );
+  fputs( "\nOptions:\n"
+    "  -h, --help                 display this help and exit\n"
+    "  -V, --version              output version information and exit\n"
+    "  -c, --stdout               write to standard output, keep input files\n"
+    "  -d, --decompress           decompress (this is the default)\n"
+    "  -f, --force                overwrite existing output files\n"
+    "  -I, --in-place             decompress or test using only one buffer\n"
+    "  -k, --keep                 keep (don't delete) input files\n"
+    "  -o, --output=<file>        write to <file>, keep input files\n"
+    "  -q, --quiet                suppress all messages\n"
+    "  -t, --test                 test compressed file integrity\n"
+    "  -v, --verbose              be verbose (a 2nd -v gives more)\n"
+    "These options are ignored when --in-place is in effect:\n"
+    "      --insize[=<size>]      pre-allocate and fill inbuf [default 16 KiB]\n"
+    "      --outsize[=<size>]     pre-allocate outbuf [default 512 MiB]\n"
+    "      --nofill               do not pass a fill function; requires --insize\n"
+    "      --noflush              do not pass a flush function; requires --outsize\n"
+    "\nIf no file names are given, or if a file is '-', xlunzip decompresses\n"
+    "from standard input to standard output.\n"
+    "Numbers may contain underscore separators between groups of digits and\n"
+    "may be followed by a SI or binary multiplier: 1_234_567kB, 4KiB.\n"
+    "\nTo extract all the files from archive 'foo.tar.lz', use the commands\n"
+    "'tar -xf foo.tar.lz' or 'xlunzip -cd foo.tar.lz | tar -xf -'.\n"
+    "\n*Exit status*\n"
+    "0 for a normal exit, 1 for environmental problems (file not found, invalid\n"
+    "command-line options, I/O errors, etc), 2 to indicate a corrupt or invalid\n"
+    "input file, 3 for an internal consistency error (e.g., bug) which caused\n"
+    "xlunzip to panic.\n"
+    "\nReport bugs to lzip-bug@nongnu.org\n"
+    "Xlunzip home page: http://www.nongnu.org/lzip/xlunzip.html\n", stdout );
   }
 
 
@@ -148,19 +150,19 @@ static void show_version( void )
   {
   printf( "%s %s\n", program_name, PROGVERSION );
   printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
-          "This is free software: you are free to change and redistribute it.\n"
-          "There is NO WARRANTY, to the extent permitted by law.\n" );
+  fputs( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
+         "This is free software: you are free to change and redistribute it.\n"
+         "There is NO WARRANTY, to the extent permitted by law.\n", stdout );
   }
 
 
 /* assure at least a minimum size for buffer 'buf' */
-static void * resize_buffer( void * buf, const unsigned min_size )
+static char * resize_buffer( void * buf, const unsigned min_size )
   {
   if( buf ) buf = realloc( buf, min_size );
   else buf = malloc( min_size );
   if( !buf ) { show_error( mem_msg, 0, false ); cleanup_and_fail( 1 ); }
-  return buf;
+  return (char *)buf;
   }
 
 
@@ -219,26 +221,72 @@ static void Pp_show_msg( Pretty_print * const pp, const char * const msg )
   }
 
 
-/* separate numbers of 5 or more digits in groups of 3 digits using '_' */
-static const char * format_num3( unsigned long long num )
+static int chvalue( const unsigned char ch )
   {
-  enum { buffers = 8, bufsize = 4 * sizeof num, n = 10 };
+  if( ch >= '0' && ch <= '9' ) return ch - '0';
+  if( ch >= 'A' && ch <= 'Z' ) return ch - 'A' + 10;
+  if( ch >= 'a' && ch <= 'z' ) return ch - 'a' + 10;
+  return 255;
+  }
+
+static unsigned long strtoul_( const char * const ptr,
+                               const char ** tail, int base )
+  {
+  if( tail ) *tail = ptr;				/* error value */
+  int i = 0;
+  while( isspace( ptr[i] ) || (unsigned char)ptr[i] == 0xA0 ) ++i;
+  const bool minus = ptr[i] == '-';
+  if( minus || ptr[i] == '+' ) ++i;
+  if( base < 0 || base > 36 || base == 1 ||
+      ( base == 0 && !isdigit( ptr[i] ) ) ||
+      ( base != 0 && chvalue( ptr[i] ) >= base ) )
+    { errno = EINVAL; return 0; }
+
+  if( base == 0 )
+    {
+    if( ptr[i] != '0' ) base = 10;			/* decimal */
+    else if( ptr[i+1] == 'x' || ptr[i+1] == 'X' ) { base = 16; i += 2; }
+    else base = 8;					/* octal or 0 */
+    }
+  const int dpg = ( base != 16 ) ? 3 : 2;	/* min digits per group */
+  int dig = dpg - 1;	/* digits in current group, first may have 1 digit */
+  unsigned long result = 0;
+  bool erange = false;
+  for( ; ptr[i]; ++i )
+    {
+    if( ptr[i] == '_' ) { if( dig < dpg ) break; else { dig = 0; continue; } }
+    const int val = chvalue( ptr[i] ); if( val >= base ) break; else ++dig;
+    if( !erange && ( ULONG_MAX - val ) / base >= result )
+      result = result * base + val;
+    else erange = true;
+    }
+  if( dig < dpg ) { errno = EINVAL; return 0; }
+  if( tail ) *tail = ptr + i;
+  if( erange ) { errno = ERANGE; return ULONG_MAX; }
+  return minus ? 0UL - result : result;
+  }
+
+
+/* separate numbers of 5 or more digits in groups of 3 digits using '_' */
+static const char * format_num3r( unsigned long long num, const bool raw )
+  {
+  enum { buffers = 8, bufsize = 4 * sizeof num };
   const char * const si_prefix = "kMGTPEZYRQ";
   const char * const binary_prefix = "KMGTPEZYRQ";
-  static char buffer[buffers][bufsize];	/* circle of static buffers for printf */
+  static char buffer[buffers][bufsize];	/* circle of buffers for printf */
   static int current = 0;
   int i;
   char * const buf = buffer[current++]; current %= buffers;
   char * p = buf + bufsize - 1;		/* fill the buffer backwards */
-  *p = 0;	/* terminator */
-  if( num > 9999 )
+  *p = 0;				/* terminator */
+  if( !raw && num >= 10000 )
     {
     char prefix = 0;			/* try binary first, then si */
-    for( i = 0; i < n && num != 0 && num % 1024 == 0; ++i )
+    for( i = 0; num != 0 && num % 1024 == 0 && binary_prefix[i]; ++i )
       { num /= 1024; prefix = binary_prefix[i]; }
     if( prefix ) *(--p) = 'i';
     else
-      for( i = 0; i < n && num != 0 && num % 1000 == 0; ++i )
+      for( i = 0; num != 0 && num % 1000 == 0 && si_prefix[i]; ++i )
         { num /= 1000; prefix = si_prefix[i]; }
     if( prefix ) *(--p) = prefix;
     }
@@ -252,6 +300,9 @@ static const char * format_num3( unsigned long long num )
   return p;
   }
 
+static const char * format_num3p( unsigned long long num )
+  { return format_num3r( num, false ); }
+
 
 void show_option_error( const char * const arg, const char * const msg,
                         const char * const option_name )
@@ -268,19 +319,19 @@ static unsigned long getnum( const char * const arg,
                              const unsigned long llimit,
                              const unsigned long ulimit )
   {
-  char * tail;
+  const char * tail;
   errno = 0;
-  unsigned long result = strtoul( arg, &tail, 0 );
+  unsigned long result = strtoul_( arg, &tail, 0 );
   if( tail == arg )
     { show_option_error( arg, "Bad or missing numerical argument in",
                          option_name ); exit( 1 ); }
 
-  if( !errno && tail[0] )
+  if( !errno && *tail )
     {
     const unsigned factor = (tail[1] == 'i') ? 1024 : 1000;
     int exponent = 0;				/* 0 = bad multiplier */
     int i;
-    switch( tail[0] )
+    switch( *tail )
       {
       case 'Q': exponent = 10; break;
       case 'R': exponent = 9; break;
@@ -308,8 +359,8 @@ static unsigned long getnum( const char * const arg,
     {
     if( verbosity >= 0 )
       fprintf( stderr, "%s: '%s': Value out of limits [%s,%s] in "
-               "option '%s'.\n", program_name, arg, format_num3( llimit ),
-               format_num3( ulimit ), option_name );
+               "option '%s'.\n", program_name, arg, format_num3p( llimit ),
+               format_num3p( ulimit ), option_name );
     exit( 1 );
     }
   return result;
@@ -357,6 +408,10 @@ static void set_d_outname( const char * const name, const int eindex )
   }
 
 
+static const char * format_num3( unsigned long long num )
+  { return format_num3r( num, true ); }
+
+
 static int open_instream( const char * const name, struct stat * const in_statsp,
                           const bool one_to_one )
   {
@@ -535,7 +590,7 @@ long flush( void * buf, unsigned long size )
   }
 
 static const char * global_name;	/* copy of filename for 'error' */
-static void error(char *x) { show_file_error( global_name, x, 0 ); }
+static void error(const char *x) { show_file_error( global_name, x, 0 ); }
 
 
 static int decompress( const int infd, Pretty_print * const pp,
@@ -543,9 +598,9 @@ static int decompress( const int infd, Pretty_print * const pp,
                        const bool nofill, const bool noflush, const bool testing )
   {
   long in_len = cl_insize;
-  uint8_t * const inbuf = (in_len > 0) ? malloc( in_len ) : 0;
+  uint8_t * const inbuf = (in_len > 0) ? (uint8_t *)malloc( in_len ) : 0;
   long out_size = cl_outsize;
-  uint8_t * const outbuf = (out_size > 0) ? malloc( out_size ) : 0;
+  uint8_t * const outbuf = (out_size > 0) ? (uint8_t *)malloc( out_size ) : 0;
   long in_pos, out_pos;
   int retval;
   if( ( in_len > 0 && !inbuf ) || ( out_size > 0 && !outbuf ) )
@@ -583,15 +638,13 @@ void show_results( Pretty_print * const pp, const long in_pos,
   if( verbosity >= 1 ) Pp_show_msg( pp, 0 );
   if( verbosity >= 2 )
     {
-    if( out_pos <= 0 || in_pos <= 0 )
-      fputs( "no data compressed. ", stderr );
-    else
-      fprintf( stderr, "%6.3f:1, %5.2f%% ratio, %5.2f%% saved. ",
+    if( verbosity >= 3 && out_pos > 0 && in_pos > 0 )
+      fprintf( stderr, "%6.3f:1, %5.2f%% ratio, %5.2f%% saved, ",
                (double)out_pos / in_pos,
                ( 100.0 * in_pos ) / out_pos,
                100.0 - ( ( 100.0 * in_pos ) / out_pos ) );
-    if( verbosity >= 3 )
-      fprintf( stderr, "%9lu out, %8lu in. ", out_pos, in_pos );
+    fprintf( stderr, "%11s out, %10s in. ",
+             format_num3( out_pos ), format_num3( in_pos ) );
     }
   if( verbosity >= 1 ) fputs( testing ? "ok\n" : "done\n", stderr );
   }
@@ -696,7 +749,8 @@ int main( const int argc, const char * const argv[] )
       case opt_insize:
         cl_insize = arg[0] ? getnum( arg, pn, 1, LONG_MAX ) : 16384; break;
       case opt_outsize: cl_outsize = arg[0] ?
-        getnum( arg, pn, min_dictionary_size, LONG_MAX ) : max_dictionary_size;
+        getnum( arg, pn, min_dictionary_size, LONG_MAX ) :
+        (unsigned)max_dictionary_size;
         break;
       case opt_nofill: nofill = true; break;
       case opt_noflush: noflush = true; break;
@@ -710,8 +764,9 @@ int main( const int argc, const char * const argv[] )
 #endif
 
   static const char ** filenames = 0;
-  int num_filenames = max( 1, ap_arguments( &parser ) - argind );
-  filenames = resize_buffer( filenames, num_filenames * sizeof filenames[0] );
+  const int num_filenames = max( 1, ap_arguments( &parser ) - argind );
+  filenames = (const char **)
+    resize_buffer( filenames, num_filenames * sizeof filenames[0] );
   filenames[0] = "-";
 
   int i;
@@ -737,7 +792,7 @@ int main( const int argc, const char * const argv[] )
   static Pretty_print pp;
   Pp_init( &pp, filenames, num_filenames );
 
-  int failed_tests = 0;
+  unsigned failed_tests = 0;
   int retval = 0;
   const bool one_to_one = !to_stdout && !testing && !to_file;
   bool stdin_used = false;
@@ -806,7 +861,7 @@ int main( const int argc, const char * const argv[] )
     set_retval( &retval, 1 );
     }
   if( failed_tests > 0 && verbosity >= 1 && num_filenames > 1 )
-    fprintf( stderr, "%s: warning: %d %s failed the test.\n",
+    fprintf( stderr, "%s: warning: %u %s failed the test.\n",
              program_name, failed_tests,
              ( failed_tests == 1 ) ? "file" : "files" );
   free( output_filename );
